// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef _SILECS_EXCEPTION_H_
#define _SILECS_EXCEPTION_H_

#include <stdexcept>
#include <stdint.h>

namespace Silecs
{

/*!
 * \brief Defines the category of the SILECS Exception.
 * Can be used in the Exception catch to dispatch emergency actions.
 * Use Exception::getCategory() to retrieve it.
 */
typedef enum
{
    CLIENT_SYSTEM_FAULT = 1,
    PLC_SYSTEM_FAULT,
    CNV_SYSTEM_FAULT,
    XML_PARSING_FAULT,
    DATA_FAULT,
    PARAM_FAULT,
    COMM_FAULT,
    CONFIG_FAULT,
    DIAG_FAULT,
    NO_FAULT,
    UNKNOWN_FAULT
} ErrorCategory;

/*!
 * \brief Defines the detailed code of the SILECS Exception.
 * Can be used in the Exception catch to execute the appropriate action.
 * Use SilecsException::getCode() to retrieve it.
 */
typedef enum
{
    XML_FILE_NOT_WELLFORMED = 1,
    XML_ELEMENT_NOT_FOUND,
    XML_EXPRESSION_MISMATCH,
    XML_CREATE_XPATH_FAULT,
    XML_EVALUATE_XPATH_FAULT,
    XML_ATTRIBUTE_NOT_FOUND,
    XML_DATA_NOT_CONSISTENT,
    XML_DATA_TYPE_MISMATCH,
    DATA_UNKNOWN_FORMAT_TYPE,
    DATA_UNKNOWN_ACCESS_TYPE,
    DATA_UNKNOWN_PLC_MANUFACTURER,
    DATA_UNKNOWN_PLC_MODEL,
    DATA_UNKNOWN_PLC_SYSTEM,
    DATA_UNKNOWN_PROTOCOL_MODE,
    DATA_READ_ACCESS_TYPE_MISMATCH,
    DATA_WRITE_ACCESS_TYPE_MISMATCH,
    DATA_ARRAY_LENGTH_MISMATCH,
    PARAM_INCORRECT_ARGUMENTS,
    PARAM_FILE_NOT_FOUND,
    PARAM_UNKNOWN_BLOCK_NAME,
    PARAM_UNKNOWN_DEVICE_NAME,
    PARAM_UNKNOWN_REGISTER_NAME,
    PARAM_ACCESS_TYPE_MISMATCH,
    PARAM_FORMAT_TYPE_MISMATCH,
    PARAM_ARRAY_DIMENSION_MISMATCH,
    PARAM_NO_TIME_STAMP,
    PARAM_INCORRECT_BLOCK_ADDRESS,
    PARAM_EXCEEDED_BLOCK_SIZE,
    PARAM_UNKNOWN_IP_ADDRESS,
    PARAM_UNKNOWN_PLC_HOSTNAME,
    COMM_CONNECT_FAILURE,
    COMM_RECONNECT_FAILURE,
    COMM_CONNECT_CLOSED,
    COMM_ALREADY_ENABLED,
    COMM_ALREADY_CONNECTED,
    COMM_NOT_CONNECTED_YET,
    COMM_BLOCK_RESIZING_NOT_ALLOWED,
    CONFIG_CLASS_NOT_DEPLOYED,
    CONFIG_CLIENT_PLC_NOT_CONSISTENT,
    CONFIG_CLIENT_PLC_NOT_AVAILABLE,
    DIAG_SLAVE_REGISTER_NOT_INITIALIZED,
    DIAG_PLC_REPORT_NOT_SUPPORTED,
    CNV_INTERNAL_ERROR,
    UNEXPECTED_ERROR,
    UNKNOWN_ERROR,
    DATA_UNKNOWN_ACCESS_AREA
} ErrorCode;

/*!
 * \class SilecsException
 * \brief This class is the based class for the Exceptions generated by the SILECS library.
 * An error belongs to a category, it has a code identifier within the category and
 * a description message.
 */
class SilecsException : public std::exception
{
public:

    /// @cond
    SilecsException(const SilecsException & ex);
    virtual ~SilecsException() throw();

    /*!
     * \brief Constructor and copy-constructors
     * \param file-name, line-number and error-code for specific faults
     * \param file-name, line-number, error-no and error-str for system faults or extra-str if needed
     */
    SilecsException(const std::string& file, const uint32_t line, const ErrorCode code, const std::string& extMsg = std::string());
    SilecsException(const std::string& file, const uint32_t line, const int err, const char *str);
    SilecsException(const std::string& file, const uint32_t line, const std::string& msg);

    /*!
     * \return a string with the error description
     */
    const char* what() const throw();

    /*!
     * \brief This method displays a warning for a not defined exception
     */
    //static void unknown(const std::string file, const long line);
    /// @endcond
    /*!
     * \brief Returns the category of the Exception object.
     * \return value from the ErrorCategory enumeration
     */
    ErrorCategory getCategory() const;

    /*!
     * \brief Returns the error code of the Exception object.
     * \return value from the ErrorCode enumeration
     */
    ErrorCode getCode() const;

    /*!
     * \brief Returns string message describing the Exception object.
     * \return Error message as string
     */
    const char* getMessage() const;

    /*!
     * \brief Print readable message of the exception to the standard output (console):
     * Time-stamp, Category, Code, Source File/Line and extra message
     */
    void print() const;

private:
    /*!
     * \fn fillCategoryMessageFromCode
     * \brief This method assign the appropriate message to a given errCode
     */
    void fillCategoryMessageFromCode(ErrorCode code);

    void buildFullMessage();

    /*!
     * \brief Log readable message of the exception to the standard output (if ERROR topic enable):
     * Time-stamp, Category, Code, Source File/Line and extra message
     */
    void log() const;

    std::string errFile_;
    long errLine_;
    ErrorCategory errCategory_;
    ErrorCode errCode_;
    std::string errMessage_;
    std::string errFullMessage_;
};

} // namespace

#endif //_SILECS_EXCEPTION_H_
