// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef _THREAD_H_
#define _THREAD_H_

#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <utility>
#include <map>
#include <queue>
#include <pthread.h>
#include <signal.h>
#include <malloc.h>
#include <string.h>

#include <silecs-communication/interface/core/Context.h>
#include <silecs-communication/interface/utility/Condition.h>

using namespace std;

namespace Silecs
{
class Context;

/*! -------------------------------------------------------------------------
 * \class Task
 * \brief Template for the Task class.
 * A Task class is defined by what the user wants to execute (user_function)
 * and by the data he wants to use (user_data).
 * It is up to the user to specify the Task template according to
 * his data type.
 */
template<typename UserDataType>
class Task
{
public:
    Task(int (*pFunc)(UserDataType, Context* pContext), UserDataType data);
    int execute(Context* pContext);

private:
    int (*userFunc)(UserDataType, Context* pContext); //function user
    UserDataType userData_; //data user
};

template<typename UserDataType>
Task<UserDataType>::Task::Task(int (*pFunc)(UserDataType, Context* pContext), UserDataType data) :
                userFunc(pFunc),
                userData_(data)
{
}

template<typename UserDataType>
int Task<UserDataType>::execute(Context* pContext)
{
    return userFunc(userData_, pContext);
}

/*! -------------------------------------------------------------------------
 * \class Thread
 * \brief Template for the Thread class.
 * One Thread instance is required per PLC connection.
 * The thread allows delegated transaction from the Cluster API level.
 * Thus, a Send/Recv Cluster call can start action on each PLC in parallel.
 */
template<typename UserDataType>
class Thread
{
public:
    Thread();
    ~Thread();
    unsigned int getPriority();
    void setPriority(unsigned int rtprio);
    void schedule(Task<UserDataType> *pt, Context* pContext);
    void waitForTaskCompletion();
private:
    static void *run(void *arg);
    void execute();
    Task<UserDataType> *pTask_;
    Context* pContext_; //to transmit data from the caller context: Cluster, PLC or Device
    pthread_t *pTid_;
    Condition* pThreadCompletion_; //condition-variable for thread loop synchronization
    bool endThread_; //boolean predicate associated with pThreadCompletion
    Condition* pTaskCompletion_; //condition-variable for task process synchronization
    bool endTask_; //boolean predicate associated with pTaskCompletion
    Mutex* pThreadBusy_; //mutex to protect thread against concurrent asynchronous call
};

template<typename UserDataType>
Thread<UserDataType>::Thread() :
                endThread_(false),
                endTask_(true)
{
    int err;
    pThreadBusy_ = new Mutex("threadBusy");
    pThreadCompletion_ = new Condition("threadCompletion");
    pTaskCompletion_ = new Condition("taskCompletion");

    pTid_ = (pthread_t *)malloc(sizeof(pthread_t));
    pThreadCompletion_->lock(); // The constructor will wait until thread is ready
    if ( (err = pthread_create(pTid_, NULL, run, this)) != 0)
        throw SilecsException(__FILE__, __LINE__, errno, strerror(err));
    //err = pthread_detach(*pTid_); not necessary because PLC thread are infinite loop
    pThreadCompletion_->wait(); // Wait until thread is ready
    pThreadCompletion_->unlock(); // Thread is ready
}
;

template<typename UserDataType>
void *Thread<UserDataType>::run(void *pThisThread)
{
    Thread *pthis = (Thread*)pThisThread;

    //An SILECS thread should not be disturbed by signals that may come from the host process.
    //Then, block signals which are generated by the system itself (means RT signals).
    sigset_t smask;
    sigemptyset(&smask);
    for (int snum = SIGRTMIN; snum <= SIGRTMAX; sigaddset(&smask, snum++))
    {
        int err;
        if ( (err = pthread_sigmask(SIG_BLOCK, &smask, NULL)) != 0)
            throw SilecsException(__FILE__, __LINE__, errno, strerror(err));
    }
    pthis->execute();
    return 0;
}
;

template<typename UserDataType>
void Thread<UserDataType>::execute()
{
    pThreadCompletion_->lock(); // The constructor will waiting until thread is ready
    pThreadCompletion_->signal(); // Thread is ready (for constructor waiting)

    while (endThread_ != true)
    {
        pThreadCompletion_->wait(); // Wait for schedule or end-Thread request
        if (endThread_ == true)
            break; // end-thread request: exit infinite loop

        try
        {
            pTaskCompletion_->lock(); // To protect the enTask_ condition variable
            //check if pThreadCompletion.wait has not been canceled by spurious signal
            if (endTask_ == false)
            { //ok, the task has to be executed
                pTask_->execute(pContext_); // Schedule request: task execution
                endTask_ = true; // Task is finished
                pTaskCompletion_->signal(); // Inform caller thread that task has been completed
            }
            pTaskCompletion_->unlock();
        }
        catch(std::exception& ex)
        {
            std::cerr << "[Thread::execute] ERROR: during execution:" << ex.what() << std::endl;
        }
        catch(...)
        {
            std::cerr << "[Thread::execute] ERROR: during execution - unknown exception received" << std::endl;
        }
    }

    pThreadCompletion_->unlock(); // Thread is terminated
}
;

template<typename UserDataType>
unsigned int Thread<UserDataType>::getPriority()
{
    int policy;
    sched_param params;

    if (pthread_getschedparam(*pTid_, &policy, &params) != 0)
    {
        string errMsg(string("Cannot get dynamically schedparams: ") + strerror(errno));
        throw errMsg;
    }
    return params.sched_priority;
}

template<typename UserDataType>
void Thread<UserDataType>::setPriority(unsigned int rtprio)
{
    int policy;
    sched_param params;

    if (pthread_getschedparam(*pTid_, &policy, &params) == 0)
    {
        params.sched_priority = rtprio;
        if (pthread_setschedparam(*pTid_, sched_getscheduler(getpid()), &params) == 0)
            return;
    }
    string errMsg(string("Cannot set dynamically schedparams: ") + strerror(errno));
    throw errMsg;
}

template<typename UserDataType>
void Thread<UserDataType>::waitForTaskCompletion()
{
    pTaskCompletion_->lock(); // To protect the enTask_ condition variable check
    while (endTask_ == false)
    {
        pTaskCompletion_->wait(); // Wait for task completion
    }
    pTaskCompletion_->unlock(); // Task has been completed
    pThreadBusy_->unlock(); // Thread has finished to execute the current task (resource is free)
}

template<typename UserDataType>
void Thread<UserDataType>::schedule(Task<UserDataType> *pt, Context* pContext)
{
    pThreadBusy_->lock(); // Thread resource must be locked until full task completion
    pThreadCompletion_->lock(); // To protect the endThread_ condition variable update
    pContext_ = pContext; // Update the caller context and
    pTask_ = pt; // the task reference to be executed
    endTask_ = false; // Task is not completed yet
    pThreadCompletion_->signal(); // Wake-up the thread to execute this referred task
    pThreadCompletion_->unlock();
}
;

template<typename UserDataType>
Thread<UserDataType>::~Thread()
{
    int err;
    pThreadCompletion_->lock(); // To protect the endThread_ condition variable
    endThread_ = true; // Thread shall stop on the next loop
    pThreadCompletion_->signal(); // Wake-up the thread to execute this referred task
    pThreadCompletion_->unlock();

    err = pthread_join(*pTid_, NULL); // Wait for thread end before continuing clean-up

    delete pTaskCompletion_;
    delete pThreadCompletion_;
    delete pThreadBusy_;

    // patch by Stefano Magnoni (20-03-2012)
    // avoid memory leak when deallocating the thread
    free(pTid_);
    pTid_ = NULL;

}
;

} // namespace

#endif // _THREAD_H_
